#ifndef _UMI_UMI_USER_HXX
#define _UMI_UMI_USER_HXX

/**
 * \file umi_user.hxx Universal Message Interface - User C++ part
 *
 * %LEGAL%
 */
#include "umic/umi.hxx"
#include "cplRaw/failable.hxx"

#ifndef EOF 
  #define EOF (-1)
#endif


namespace INdepUmi100 {

  /**
   * Universal Message Interface -
   * UMI interface definition
   *
   * \note 'Failure' in the context of a Umi always means 'transportation' 
   * never 'endpoint' failure.
   */
  class Umi:public Failable {
    const UmiProto *_pup;

   public:
    const UmiProto *pup() { return _pup; }
    /*
     * transport failures
     */
    enum {
     OK,
     E_NOBUF
    };

    /**
     * send a message downstream.
     * \param puc Pointer to the control part of the message
     * \param pud Pointer to the data part of the message
     * \return (Failable&) *this
     */
    virtual Failable& putMsg(const UmiControl*puc, const UmiData*pud) = 0;

    /**
     * get a upstream message
     * \param puc Pointer to the control part buffer
     * \param pud Pointer to the data part buffer
     * \return (Failable&) *this
     */
    virtual Failable& getMsg(UmiControlBuf *puc, UmiData *pud) = 0;

    /**
     * exchange message
     * \param pucd Pointer to the downstream control part of the message
     * \param pudd Pointer to the downstream data part of the message
     * \param pucu Pointer to the upstream control part buffer
     * \param pudu Pointer to the upstream data part buffer
     * \return (Failable&) *this
     */
    virtual Failable& xchMsg(const UmiControl*pucd, const UmiData*pudd, UmiControlBuf *pucu, UmiData *pudu);

    /* flow control prevent deadlock */
    virtual bool canPut() ;


    /* data max per message */
    size_t sDataMax() { return __SZ_DATA_MAX; }		

    Umi(const UmiProto *puf);
    virtual ~Umi();
  };



  /**
   * Factory interface definition.
   *
   * Using a Factory, UMI channels can be constructed without
   * knowledge of the specific UMI in use. All channels for a
   * specific protocol must be created using the same Factory.
   * 
   * \note In order to assist factories which keep state between the
   * creation of channels, all channels must be created in increasing channel 
   * order ( 0, 1, ... ).
   * Furthermore, the Umi channel objects created by createUmi() must be deleted
   * in reverse order ( ... , 1 , 0 ).
   */
  class Factory {
   public:
    /**
     * create channel to driver specified by protocol and channel number
     * (thread-safe)
     * \param ppum Pointer to Umi*, returns the UMI created. Set to 0 if the creation did fail.
     * \param pup Pointer to the UMI protocol to be spoken on the channel.
     * \param iChan channel specification.
     * \return A Failable indicating the state of the created Umi, even if this has not been created and is not returned.
     */
    virtual Failable createUmi(Umi**ppum, const UmiProto *pup, int iChan) =0;

    Factory() {} 
    virtual ~Factory() {}
  };

  /** factory maker */
  class FactoryImplementation {
    const char *_szName;
   public:
    const char *szName() { return _szName; }
    virtual Factory* createFactory(const char *szMode, const char *szArg)=0;
    FactoryImplementation(const char *szName):_szName(szName) {}
    virtual ~FactoryImplementation() {}
  };

  /**
   * implementation dependent list of Factories
   * (thread-safe)
   */
  class Factories {

    /** 
     * glocal factory table, to be set up by global initializers
     */
    static FactoryImplementation* mpPff[4];

   public:
    void registerFactoryImplementation(FactoryImplementation *);
    
    /** 
     * construct Factory from internal list
     */
    Factory* createFactory(const char *szArg);

    Factories();
     ~Factories();
  };

  /**
   * implemented factories (global)
   */
  extern Factories fact;
}
#endif


